package v3d.editor;

import java.applet.*;
import java.awt.*;
import java.io.*;
import java.net.URL;

/*
 * AppletFileIO handles the IO between the editor and the database/file. Only constructor,
 * start and getGUI are normally called from the editor applet/application. The GUI
 * handles the rest of the calls, such as load, save or upload.
 *
 * AppletFileIO can be configured by giving parameters to the applet. This way the page
 * containing the applet can be generated using the database, and normal URL mechanisms
 * can be used for fetching the content from the database based on that. The same
 * mechanism can be used for uploading the files back to the database, so that the
 * upload URL's are generated to the page contaning the applet.
 *
 * Currently the parameters read from the editor applet are "scene" (the name of the
 * scene file), "userMessage" (the name of the user message string file), "editingInfo"
 * (the info string about current editing session) and "exitURL", the URL to which
 * the applet exits after upload.
 */
class AppletFileIO extends FileIO {

	public AppletFileIO(Editor editor) {
		super(editor);
		ioPanel = new AppletFileIOPanel(this);
	}

	// starts the FileIO from editor start function
	public void start() {
		String sceneName = editor.getParameter("scene");
//		System.out.println("scene = " + sceneName);
		load(sceneName);

		String editingInfo = editor.getParameter("editingInfo");
//		System.out.println("editor = " + editingInfo);
		if(editingInfo != null) {
			editor.setCurrentEditor(editingInfo);
		}
		else {
			editor.setCurrentEditor("Unknown editor");
		}

		String userMessage = editor.getParameter("userMessage");
//		System.out.println("message = " + userMessage);
		String msg = loadString(userMessage);
		editor.setUserMsg(msg);
	}

	public void load(String sceneName) {
		if (sceneName != null) {
			Scene scene = editor.getScene();
			InputStream is = null;
			try {
				is = new URL(editor.getDocumentBase(), sceneName).openStream();
				scene.deserialize(is);
			}
			catch(Exception e) {
				scene.reset();
//				System.out.println("Error: "+e.toString());
				QuerySpawner query = new QuerySpawner();
				query.doQuery("",e.toString(),"Ok",null,null);
			}

			try {
				if (is != null)
				is.close();
			}
			catch(Exception e) {
			}
		}
	}

	public String loadString(String name) {
		if (name != null) {
			String s = new String();

			InputStream is = null;
			try {
				is = new URL(editor.getDocumentBase(), name).openStream();
				DataInputStream dis = new DataInputStream(is);
				// maybe we should rather use Unicode than 8-bit charset
				while (dis.available() > 0)
					s += dis.readLine();
			}
			catch(Exception e) {
//				System.out.println("Error: "+e.toString());
				QuerySpawner query = new QuerySpawner();
				query.doQuery("",e.toString(),"Ok",null,null);
			}

			try {
				if (is != null)
				is.close();
			}
			catch(Exception e) {
			}

			return s;
		}
		return null;
	}

	public void save() {
		Scene sc = editor.getScene();

		// Fetch the scene in native 3d6b format. Currently the format is
		// string-based, but it may change to binary to save bandwith and space
		String scene;
		scene = sc.serialize();
		// Outputs to java console, change to real save operation
		System.out.println(scene);

		// Fetch the scene in VRML format.
		String vrml;
		vrml = sc.getVRML();
		// Outputs to java console, change to real save operation
		System.out.println(vrml);

		// Fetch the scene in VRML format, but using history for the colours
		String historyVrml;
		historyVrml = sc.getHistoryVRML();
		// Outputs to java console, change to real save operation
		System.out.println(historyVrml);

		// Fetch the thumbnail image of the scene in Targa format. The format
		// has no compression, but 100*50 pixel image takes only about 15 k.
		// The image can be converted to jpeg on the server side.
		byte tga[];
		tga = editor.getThumbnailTGA(100,50,7);
		// Output the buffer somewhere
/*		try {
			FileOutputStream out = new FileOutputStream("test.tga");
			out.write(tga);
		}
		catch(Exception e) {
//			System.out.println("Error: "+e.toString());
			QuerySpawner query = new QuerySpawner();
			query.doQuery("",e.toString(),"Ok",null,null);
		}*/

		// Fetch the user message from the editor.
		String userMsg;
		userMsg = editor.getUserMsg();
		// Outputs to java console, change to real save operation
		System.out.println(userMsg);

		// Jump out from the applet

		String exitLink = editor.getParameter("exitURL");
		if (exitLink != null) {
			try {
				URL exitURL = new URL(editor.getDocumentBase(), exitLink);
				AppletContext context = editor.getAppletContext();
				context.showDocument(exitURL,"_self");
			}
			catch(Exception e) {
			}
		}

	}

}

/*
 * This is the UI panel for the FileIO. It is incorporated to the rest of the
 * user interface, but it is isolated here so that it stays as its own module
 * and can be easily changed.
 */
class AppletFileIOPanel extends Panel {
	AppletFileIO io;

	AppletFileIOPanel(AppletFileIO io)
	{
		this.io = io;
		this.setLayout(new GridLayout(1,1,2,2));
		this.add(new Button("Upload"));
	}

	public boolean action(Event e, Object arg) {
		if (e.target instanceof Button) {
			String buttonText = (String)arg;
			if (buttonText.equals("Upload")) {
				io.save();
			}
		}
		return true;
	}
}
