package v3d.editor;

import java.awt.*;
import java.io.*;

// Grid class
// - draw the grid (from a given camera)
// - rotate the grid
// - move the grid
// - bend the grid
// - enable adjustable stepping for operations above
// - get an intersection of the grid and a ray
// - enable/disable snapping
// - set the size of the grid cell
// - serialize/deserialize grid parameters
class Grid {
	int cornerIds[][];
	int cellCount;
	float cellSize;

	public LineBuffer originalLines;
	public LineBuffer tempLines;
	public VertexBuffer originalBuffer;
	public VertexBuffer transformedBuffer;
	public VertexBuffer projectedBuffer;

	Matrix4x4 transform;
	Matrix4x4 invTransform;

	float xBend, zBend;

	int gridColor;

	float stepSize;
	boolean stepEnabled;
	boolean snapEnabled;

	Vector3D stepMoveCumulation;
	Vector3D stepRotateCumulation;

	float angleMultiplier = 0.01f;
	float bendScale = 0.4f;
	float bendMultiplier = 0.00005f;
	float moveMultiplier = 0.5f;

	UndoBuffer undo = null;
/*	 _ _ _
	|_|_|_|
	|_|_|_|
	|_|_|_|
*/
	Grid() {
		this(1,1.0f,0);
	}

	Grid(int count, float size, int color) {
		cellCount = count;
		cellSize = size;

		transform = new Matrix4x4();
		invTransform = new Matrix4x4();
		
		int lines = 2*(cellCount+1)*cellCount;
		originalLines = new LineBuffer(lines+1);
		tempLines = new LineBuffer(lines+1);

		int vertices = (cellCount+1)*(cellCount+1);
		originalBuffer = new VertexBuffer(vertices+1);
		transformedBuffer = new VertexBuffer(vertices+1);
		projectedBuffer = new VertexBuffer(vertices+1);

		gridColor = color;

		xBend = 0.0f;
		zBend = 0.0f;

		stepSize = 15.0f;
		stepEnabled = false;
		snapEnabled = false;

		stepMoveCumulation = new Vector3D(0.0f,0.0f,0.0f);
		stepRotateCumulation = new Vector3D(0.0f,0.0f,0.0f);

		initGrid();
		reset();
	}

	public void initGrid() {
		cornerIds = new int[cellCount+1][cellCount+1];

		float cellCorner = -cellSize * cellCount / 2.0f;

		Vector3D tVec = new Vector3D();
		tVec.y = 0;
		int x,z;
		for(z=0; z<=cellCount; z++) {
			tVec.z = cellCorner + z*cellSize;
			for(x=0; x<=cellCount; x++) {
				tVec.x = cellCorner + x*cellSize;
				cornerIds[z][x] = originalBuffer.addVectorWithReference(tVec);
			}
		}

		for(z=0; z<=cellCount; z++) {
			for(x=0; x<cellCount-1; x++) {
				originalLines.directLine(cornerIds[z][x],cornerIds[z][x+1],gridColor,-1);
				originalBuffer.addReference(cornerIds[z][x+1]);
			}
			originalLines.directLine(cornerIds[z][cellCount-1],cornerIds[z][cellCount],gridColor,-1);
		}

		for(x=0; x<=cellCount; x++) {
			for(z=0; z<cellCount; z++) {
				originalLines.directLine(cornerIds[z][x],cornerIds[z+1][x],gridColor,-1);
				originalBuffer.addReference(cornerIds[z][x]);
				originalBuffer.addReference(cornerIds[z+1][x]);
			}
		}

		originalLines.setZOffset(100.0f);
	}

	public void reset() {
		storeUndo();

		transform.makeIdentity();
		invTransform.makeIdentity();
		zBend = 0;
		xBend = 0;
		stepSize = 15.0f;
		stepEnabled = false;
		snapEnabled = false;
		setSize(15.0f);

		storeRedo();
	}

	public void resetStepCumulation()
	{
		stepMoveCumulation.x = 0.0f;
		stepMoveCumulation.y = 0.0f;
		stepMoveCumulation.z = 0.0f;
		stepRotateCumulation.x = 0.0f;
		stepRotateCumulation.y = 0.0f;
		stepRotateCumulation.z = 0.0f;
	}

	void stepCumulate(Vector3D vec, Vector3D stepCumulation, boolean forceStepping) {
		if (stepEnabled) {
			if (forceStepping) {
				if (vec.x > 0.0f) {
					vec.x = stepSize;
				}
				else if (vec.x < 0.0f) {
					vec.x = -stepSize;
				}
				if (vec.y > 0.0f) {
					vec.y = stepSize;
				}
				else if (vec.y < 0.0f) {
					vec.y = -stepSize;
				}
				if (vec.z > 0.0f) {
					vec.z = stepSize;
				}
				else if (vec.z < 0.0f) {
					vec.z = -stepSize;
				}
			}
			else {
				stepCumulation.addToThis(vec);
				if (vec.x > 0.0f) {
					vec.x = 0.0f;
					while (stepCumulation.x > stepSize) {
						stepCumulation.x -= stepSize;
						vec.x += stepSize;
					}
				}
				else {
					vec.x = 0.0f;
					while (stepCumulation.x < -stepSize) {
						stepCumulation.x += stepSize;
						vec.x -= stepSize;
					}
				}
				if (vec.y > 0.0f) {
					vec.y = 0.0f;
					while (stepCumulation.y > stepSize) {
						stepCumulation.y -= stepSize;
						vec.y += stepSize;
					}
				}
				else {
					vec.y = 0.0f;
					while (stepCumulation.y < -stepSize) {
						stepCumulation.y += stepSize;
						vec.y -= stepSize;
					}
				}
				if (vec.z > 0.0f) {
					vec.z = 0.0f;
					while (stepCumulation.z > stepSize) {
						stepCumulation.z -= stepSize;
						vec.z += stepSize;
					}
				}
				else {
					vec.z = 0.0f;
					while (stepCumulation.z < -stepSize) {
						stepCumulation.z += stepSize;
						vec.z -= stepSize;
					}
				}
			}
		}
	}

	public void move(Vector3D vec, boolean forceStepping) {
		storeUndo();

		vec.multiplyThis(moveMultiplier);
		stepCumulate(vec,stepMoveCumulation,forceStepping);
		transform.translate(vec);
		transform.normalize();
		invTransform.copy(transform);
		invTransform.inverseFast();

		storeRedo();
	}

	public void move(Vector3D vec, Camera cam, boolean forceStepping) {
		storeUndo();

		vec.multiplyThis(moveMultiplier);
		stepCumulate(vec,stepMoveCumulation,forceStepping);
		Matrix4x4 tmat = new Matrix4x4();
		tmat.makeTranslation(vec);
		Matrix4x4 viewMatrix = new Matrix4x4(cam.viewMatrix);

		viewMatrix.values[3][0] = 0.0f;
		viewMatrix.values[3][1] = 0.0f;
		viewMatrix.values[3][2] = 0.0f;

		transform = transform.multiply(viewMatrix);
		transform = transform.multiply(tmat);
		viewMatrix.inverseFast();
		transform = transform.multiply(viewMatrix);

		transform.normalize();
		invTransform.copy(transform);
		invTransform.inverseFast();

		storeRedo();
	}

	public void rotateX(float angle, boolean forceStepping) {
		storeUndo();

		Vector3D angleVector = new Vector3D(angle * angleMultiplier * 180.0f / (float)(Math.PI),0.0f,0.0f);
		stepCumulate(angleVector,stepRotateCumulation,forceStepping);
		transform.rotateX((float)(Math.PI) * angleVector.x / 180.0f);
		transform.normalize();
		invTransform.copy(transform);
		invTransform.inverseFast();

		storeRedo();
	}

	public void rotateX(float angle, Camera cam, boolean forceStepping) {
		storeUndo();

		Vector3D angleVector = new Vector3D(angle * angleMultiplier * 180.0f / (float)(Math.PI),0.0f,0.0f);
		stepCumulate(angleVector,stepRotateCumulation,forceStepping);
		Matrix4x4 tmat = new Matrix4x4();
		tmat.makeRotationX((float)(Math.PI) * angleVector.x / 180.0f);

		Matrix4x4 viewMatrix = new Matrix4x4(cam.viewMatrix);

		viewMatrix.values[3][0] = 0.0f;
		viewMatrix.values[3][1] = 0.0f;
		viewMatrix.values[3][2] = 0.0f;

		transform = transform.multiply(viewMatrix);
		transform = transform.multiply(tmat);
		viewMatrix.inverseFast();
		transform = transform.multiply(viewMatrix);

		transform.normalize();
		invTransform.copy(transform);
		invTransform.inverseFast();

		storeRedo();
	}

	public void rotateY(float angle, boolean forceStepping) {
		storeUndo();

		Vector3D angleVector = new Vector3D(0.0f,angle * angleMultiplier * 180.0f / (float)(Math.PI),0.0f);
		stepCumulate(angleVector,stepRotateCumulation,forceStepping);
		transform.rotateY((float)(Math.PI) * angleVector.y / 180.0f);
		transform.normalize();
		invTransform.copy(transform);
		invTransform.inverseFast();

		storeRedo();
	}

	public void rotateY(float angle, Camera cam, boolean forceStepping) {
		storeUndo();

		Vector3D angleVector = new Vector3D(0.0f,angle * angleMultiplier * 180.0f / (float)(Math.PI),0.0f);
		stepCumulate(angleVector,stepRotateCumulation,forceStepping);
		Matrix4x4 tmat = new Matrix4x4();
		tmat.makeRotationY((float)(Math.PI) * angleVector.y / 180.0f);

		Matrix4x4 viewMatrix = new Matrix4x4(cam.viewMatrix);

		viewMatrix.values[3][0] = 0.0f;
		viewMatrix.values[3][1] = 0.0f;
		viewMatrix.values[3][2] = 0.0f;

		transform = transform.multiply(viewMatrix);
		transform = transform.multiply(tmat);
		viewMatrix.inverseFast();
		transform = transform.multiply(viewMatrix);

		transform.normalize();
		invTransform.copy(transform);
		invTransform.inverseFast();

		storeRedo();
	}

	public void rotateZ(float angle, boolean forceStepping) {
		storeUndo();

		Vector3D angleVector = new Vector3D(0.0f,0.0f,angle * angleMultiplier * 180.0f / (float)(Math.PI));
		stepCumulate(angleVector,stepRotateCumulation,forceStepping);
		transform.rotateZ((float)(Math.PI) * angleVector.z / 180.0f);
		transform.normalize();
		invTransform.copy(transform);
		invTransform.inverseFast();

		storeRedo();
	}

	public void rotateZ(float angle, Camera cam, boolean forceStepping) {
		storeUndo();

		Vector3D angleVector = new Vector3D(0.0f,0.0f,angle * angleMultiplier * 180.0f / (float)(Math.PI));
		stepCumulate(angleVector,stepRotateCumulation,forceStepping);
		Matrix4x4 tmat = new Matrix4x4();
		tmat.makeRotationZ((float)(Math.PI) * angleVector.z / 180.0f);

		Matrix4x4 viewMatrix = new Matrix4x4(cam.viewMatrix);

		viewMatrix.values[3][0] = 0.0f;
		viewMatrix.values[3][1] = 0.0f;
		viewMatrix.values[3][2] = 0.0f;

		transform = transform.multiply(viewMatrix);
		transform = transform.multiply(tmat);
		viewMatrix.inverseFast();
		transform = transform.multiply(viewMatrix);

		transform.normalize();
		invTransform.copy(transform);
		invTransform.inverseFast();

		storeRedo();
	}

	public void bendX(float bendChange, boolean forceStepping) {
		storeUndo();

		bendChange *= bendScale;
		Vector3D bendVector = new Vector3D(bendChange,0.0f,0.0f);
		stepCumulate(bendVector,stepMoveCumulation,forceStepping);
		xBend += bendMultiplier * bendVector.x;
		calculateBend();

		storeRedo();
	}

	public void bendZ(float bendChange, boolean forceStepping) {
		storeUndo();

		bendChange *= bendScale;
		Vector3D bendVector = new Vector3D(0.0f,0.0f,bendChange);
		stepCumulate(bendVector,stepMoveCumulation,forceStepping);
		zBend += bendMultiplier * bendVector.z;
		calculateBend();

		storeRedo();
	}

	void calculateBend() {
		// y = a1*x^2 + a2*z^2
		int xi,zi;
		float x,z;
		for(zi=0; zi<=cellCount; zi++) {
			for(xi=0; xi<=cellCount; xi++) {
				z = originalBuffer.vertices[cornerIds[zi][xi]].vec.z;
				x = originalBuffer.vertices[cornerIds[zi][xi]].vec.x;
				originalBuffer.vertices[cornerIds[zi][xi]].vec.y = xBend * x * x + zBend * z * z;
			}
		}
	}

	public void setSize(float size) {
		storeUndo();

		cellSize = size;

		originalLines.clear();
		tempLines.clear();
		originalBuffer.clear();
		transformedBuffer.clear();
		projectedBuffer.clear();

		initGrid();
		calculateBend();

		storeRedo();
	}

	public float getSize() {
		return cellSize;
	}

	public void setStepping(boolean stepping) {
		stepEnabled = stepping;
	}

	public boolean getStepping() {
		return stepEnabled;
	}

	public void setStepSize(float size) {
		stepSize = size;
	}

	public float getStepSize() {
		return stepSize;
	}

	public void setSnapping(boolean snapping) {
		snapEnabled = snapping;
	}

	public boolean getSnapping() {
		return snapEnabled;
	}

	public void setUndoBuffer(UndoBuffer ub) {
		undo = ub;
	}

	void storeUndo() {
		if (undo != null) {
			GridUndoCmd cmd = new GridUndoCmd(this);
			storeState(cmd.state);
			undo.storeToQueue(cmd);
		}
	}

	void storeRedo() {
		if (undo != null) {
			GridRedoCmd cmd = new GridRedoCmd(this);
			storeState(cmd.state);
			undo.storeToQueue(cmd);
		}
	}

	void storeState(GridState state) {
		state.transform.copy(transform);
		state.xBend = xBend;
		state.zBend = zBend;
		state.cellSize = cellSize;
	}

	public void restoreState(GridState state) {
		boolean bendChange = false;
		boolean sizeChange = false;

		transform.copy(state.transform);
		invTransform.copy(transform);
		invTransform.inverseFast();

		if (xBend != state.xBend || zBend != state.zBend)
			bendChange = true;

		xBend = state.xBend;
		zBend = state.zBend;

		if (cellSize != state.cellSize)
			sizeChange = true;

		cellSize = state.cellSize;

		if (sizeChange) {
			originalLines.clear();
			tempLines.clear();
			originalBuffer.clear();
			transformedBuffer.clear();
			projectedBuffer.clear();

			initGrid();
		}

		if (sizeChange || bendChange)
			calculateBend();
	}

	public void snapToGrid(Vector3D vec) {
		if (snapEnabled) {
			int xp = cellCount/2 + (int)Math.floor((vec.x / cellSize) + 0.5f);
			int zp = cellCount/2 + (int)Math.floor((vec.z / cellSize) + 0.5f);
			if (xp < 0)
				xp = 0;
			else if (xp > cellCount)
				xp = cellCount;

			if (zp < 0)
				zp = 0;
			else if (zp > cellCount)
				zp = cellCount;

			vec.copy(originalBuffer.vertices[cornerIds[zp][xp]].vec);
		}
	}

	public boolean getIntersection(LineSegment ray, Vector3D result)
	{
		// y = ax^2 + bz^2
		// y = cx + d
		// y = ez + f

		// y1 = cx1 + d
		// y1 = ez1 + f
		// y2 = cx2 + d
		// y2 = ez2 + f

		// y1 - cx1 = d
		// y2 -cx2 = d
		// y1 - cx1 = y2 - cx2
		// y1 - y2 = c(x1 -x2)
		// c = (y1 - y2)/(x1 - x2)
		// y1 - ez1 = f
		// y2 - ez2 = f
		// y1 - ez1 = y2 - ez2
		// y1 - y2 = e(z1 - z2)
		// e = (y1 - y2)/(z1 - z2)
		// d = y1 - c*x1
		// f = y1 - e*z1

		// y = ax^2 + bz^2
		// y = cx + d
		// y = ez + f

		// cx = y - d
		// x = (y - d)/c
		// ez = y - f
		// z = (y - f)/e

		// y = a((y - d)/c)^2 + b((y - f)/e)^2
		// y = a((y - d)^2/c^2) + b((y - f)^2/e^2
		// y = a((y^2 - 2yd + d^2)/c^2 + b((y^2 - 2yf + f^2)/e^2
		// y = y^2 (a/c^2) - y (2ad/c^2) + ad^2/c^2 + y^2 (b/e^2) - y (2bf/e^2) + bf^2/e^2

		// y^2 (a/c^2 + b/e^2) - y (2ad/c^2 + 2bf/e^2 + 1) + ad^2/c^2 + bf^2/e^2 = 0

		// x = (y - d)/c
		// x = (y - y1 + c*x1)/c
		// x = (y - y1)/c + x1
		// z = (y - f)/e
		// z = (y - y1 + e*z1)/e
		// z = (y - y1)/e + z1

		// y = a((y - y1)/c + x1)^2 + b((y - y1)/e + z1)^2
		// y = a(((y - y1)/c)^2 + x1(y - y1)/c + x1^2) + b(((y - y1)/e)^2 + z1(y - y1)/e + z1^2)
		// y = a((y^2 - y1y + y1^2)/c^2 + x1(y - y1)/c + x1^2)
		//   + b((y^2 - y1y + y1^2)/e^2 + z1(y - y1)/e + z1^2)
		// y = a(y^2/c^2 - y1y/c^2 + y1^2/c^2 + x1y/c - x1y1/c + x1^2)
		//   + b(y^2/e^2 - y1y/e^2 + y1^2/e^2 + z1y/e - z1y1/e + z1^2)
		// y = ay^2/c^2 - ay1y/c^2 + ay1^2/c^2 + ax1y/c - ax1y1/c + ax1^2
		//   + by^2/e^2 - by1y/e^2 + by1^2/e^2 + bz1y/e - bz1y1/e + bz1^2
		// y = y^2 * (a/c^2 + b/e^2) + y (- ay1/c^2 + ax1/c - by1/e^2 + bz1/e)
		//   + ay1^2/c^2 - ax1y1/c + ax1^2 + by1^2/e^2 - bz1y1/e + bz1^2
		//   y^2 * (a/c^2 + b/e^2)
		// + y (- ay1/c^2 + ax1/c - by1/e^2 + bz1/e - 1)
		// + ay1^2/c^2 - ax1y1/c + ax1^2 + by1^2/e^2 - bz1y1/e + bz1^2 = 0

		// solution for second order equation of form
		// ax^2 + bx + c = 0
		// is
		// D = b^2 - 4ac
		// x = (- b +- sqrt(D))/2a

		// if D > 0 fuction has two solutions
		// if D = 0 function has one solution
		// if D < 0 function has no solution (only imaginary roots)

		// After solutions are found, a range check with the grid should be performed,
		// and the solution closer to the start of the ray should be selected

		Vector3D start = new Vector3D();
		Vector3D end = new Vector3D();

		invTransform.transform(ray.start,start);
		invTransform.transform(ray.end,end);

		float c = (start.y - end.y)/(start.x - end.x);
		float d = start.y - c*start.x;
		float e = (start.y - end.y)/(start.z - end.z);
		float f = start.y - e*start.z;

		float ic = (start.x - end.x)/(start.y - end.y);
		float ie = (start.z - end.z)/(start.y - end.y);

		float a = xBend / (c * c) + zBend / (e * e);
		float b = -(2.0f * xBend * d / (c * c) + 2.0f * zBend * f / (e * e) + 1.0f);
		float cc = xBend * d * d / (c * c) + zBend * f * f / (e * e); 

		//   y^2 * (a/c^2 + b/e^2)
		// + y (- ay1/c^2 + ax1/c - by1/e^2 + bz1/e - 1)
		// + ay1^2/c^2 - ax1y1/c + ax1^2 + by1^2/e^2 - bz1y1/e + bz1^2 = 0

/*		float a = xBend * (ic * ic) + zBend * (ie * ie);
		float b = -xBend * start.y * (ic * ic) + xBend * start.x * ic
				- zBend * start.y * (ie * ie) + zBend * start.z * ie - 1.0f;
		float cc = xBend * ( start.y * start.y * (ic * ic) - start.x * start.y * ic + start.x * start.x)
				+ zBend * (start.y * start.y * (ie * ie) - start.z * start.y * ie + start.z * start.z);
*/
		float dis = b * b - 4.0f * a * cc;

//		System.out.println("a: " + a + " b: " + b + " c: " + c + " D: " + d);

		int solutions;
		Vector3D vec[];
		vec = new Vector3D[2];
		if (dis > 0) {
			solutions = 2;
			vec[0] = new Vector3D();
			vec[1] = new Vector3D();
			float sqd = (float)Math.sqrt(dis);
			if ( a != 0 ) {
				vec[0].y = (-b + sqd) / (2 * a);
				vec[1].y = (-b - sqd) / (2 * a);
			}
			else {
				vec[0].y = 0;
				vec[1].y = 0;
			}
		}
		else if (dis == 0) {
			solutions = 1;
			vec[0] = new Vector3D();
			if ( a != 0 ) {
				vec[0].y = - b / (2 * a);
			}
			else {
				vec[0].y = 0;
			}
		}
		else {
			return false;
		}

		int i;
		float cellCorner = cellSize * cellCount / 2.0f;
		int inRange = 0;
		int rangeIndex = 0;
		for (i=0; i<solutions; i++) {
			vec[i].x = (vec[i].y - start.y)/c + start.x;
			vec[i].z = (vec[i].y - start.y)/e + start.z;
			if (vec[i].x >= -cellCorner && vec[i].x <= cellCorner &&
				vec[i].z >= -cellCorner && vec[i].z <= cellCorner) {
				inRange++;
				rangeIndex = i;
			}
		}

		if (inRange == 0)
			return false;
		else if (inRange == 1) {
			snapToGrid(vec[rangeIndex]);
			transform.transform(vec[rangeIndex],result);
			return true;
		}
		else {
			Vector3D diff = vec[0].subtract(start);
			float len1 = diff.length();
			diff = vec[1].subtract(start);
			float len2 = diff.length();
			if (len1 < len2) {
				snapToGrid(vec[0]);
				transform.transform(vec[0],result);
			}
			else {
				snapToGrid(vec[1]);
				transform.transform(vec[1],result);
			}
			return true;
		}
	}

	public void prepareForDraw(Camera cam, LineBuffer targetLines, VertexBuffer targetVertices) {
		Matrix4x4 tempMatrix = transform.multiply(cam.viewMatrix);

		originalBuffer.transform(tempMatrix,transformedBuffer);

		tempLines.copy(originalLines);
		tempLines.zClip(transformedBuffer,cam.minZ,cam.maxZ);

		transformedBuffer.project(cam,projectedBuffer);

		tempLines.xClip(projectedBuffer,cam.viewXOffset,cam.viewXOffset+cam.viewWidth-1);
		tempLines.yClip(projectedBuffer,cam.viewYOffset,cam.viewYOffset+cam.viewHeight-1);

		tempLines.makeFog(projectedBuffer,cam.getProjectedZPlane(cam.minZ),cam.getProjectedZPlane(cam.maxZ));

		tempLines.join(projectedBuffer, targetLines, targetVertices);
	}

	public void print() {
		System.out.println("cellCount: " + cellCount +
					" cellSize: " + cellSize +
					" xBend: " + xBend +
					" zBend: " + zBend +
					" gridColor: " + gridColor);
	}

	// Grid is defined with a line starting with word "grid". The first 9 values on the
	// line are similar to the camera as they define the position and orientation of the
	// grid: position (three floats), front direction (three floats) and up vector (three
	// floats). The next values are the bend values of the grid (two floats), amount of
	// cells in the grid in one direction (one integer), the size of one cell (one float)
	// and the colour of the grid (three integers).

	// Creates a string of the grid
	public String serialize() {
		String s = new String();

		Vector3D posVec = new Vector3D();
		transform.transform(new Vector3D(0,0,0), posVec);

		Vector3D frontVec = new Vector3D();
		transform.transform(new Vector3D(0,0,1), frontVec);

		Vector3D upVec = new Vector3D();
		transform.transform(new Vector3D(0,1,0), upVec);

		frontVec.subtractFromThis(posVec);
		upVec.subtractFromThis(posVec);

		s =	"grid " +
			posVec.x + " " +
			posVec.y + " " +
			posVec.z + " " +
			frontVec.x + " " +
			frontVec.y + " " +
			frontVec.z + " " +
			upVec.x + " " +
			upVec.y + " " +
			upVec.z + " " +
			xBend + " " +
			zBend + " " +
			cellCount + " " +
			cellSize + " " +
			((gridColor & 0xff0000) >> 16) + " " +
			((gridColor & 0x00ff00) >> 8) + " " +
			(gridColor & 0x0000ff) + "\n";

		return s;
	}

	// Tries to read a grid from the stream. Returns
	// true if the token was correct and values were read,
	// otherwise false.
	public boolean deserialize(StreamTokenizer st) throws IOException {
		if ("grid".equals(st.sval)) {
			Vector3D posVec, frontVec, upVec;
			posVec = new Vector3D();
			frontVec = new Vector3D();
			upVec = new Vector3D();

			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				posVec.x = (float)st.nval;
			}
			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				posVec.y = (float)st.nval;
			}
			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				posVec.z = (float)st.nval;
			}
			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				frontVec.x = (float)st.nval;
			}
			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				frontVec.y = (float)st.nval;
			}
			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				frontVec.z = (float)st.nval;
			}
			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				upVec.x = (float)st.nval;
			}
			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				upVec.y = (float)st.nval;
			}
			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				upVec.z = (float)st.nval;
			}

			transform.makeLookAt(posVec,frontVec.subtract(posVec),upVec);
			invTransform.copy(transform);
			invTransform.inverseFast();

			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				this.xBend = (float)st.nval;
			}
			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				this.zBend = (float)st.nval;
			}
		
			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				this.cellCount = (int)st.nval;
			}
			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				this.cellSize = (float)st.nval;
			}

			int r = 0, g = 0, b = 0;
			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				r = (int)st.nval;
				if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
					g = (int)st.nval;
					if (st.nextToken() == StreamTokenizer.TT_NUMBER)
						b = (int)st.nval;
				}
			}
			this.gridColor = (r << 16) | (g << 8) | b;

			while (st.ttype != StreamTokenizer.TT_EOL &&
				st.ttype != StreamTokenizer.TT_EOF)
				st.nextToken();

			int lines = 2*(cellCount+1)*cellCount;
			originalLines = new LineBuffer(lines+1);
			tempLines = new LineBuffer(lines+1);

			int vertices = (cellCount+1)*(cellCount+1);
			originalBuffer = new VertexBuffer(vertices+1);
			transformedBuffer = new VertexBuffer(vertices+1);
			projectedBuffer = new VertexBuffer(vertices+1);

			initGrid();
			calculateBend();

//			originalLines.print();
//			originalBuffer.print();
			return true;
		}
		else {
			return false;
		}
	}
}
