package v3d.editor;

import java.util.Vector;

/*
 * UndoBuffer contains all commands that are performed in VertexBuffer and LineBuffer.
 * Before each drawing operation, openQueue() is called, and after the operation,
 * closeQueue() is called. This create a block of commands in undoBuffer. When a command
 * is undone, undo() is called, and when redone, redo() is called. UndoBuffer holds
 * internally the current block of commands, so calling openQueue() after undoing
 * operations will discard all the commandblocks that are after the current block in
 * the buffer.
 */
class UndoBuffer {
	UndoQueue queues[];
	int queueArraySize;
	int firstQueueIndex;	// index of the first queue slot
	int lastQueue;		// offset from firstQueue to last queue slot
	int currentQueue;		// offset from firstQueue to current queue slot
	int currentQueueIndex;	// index of the current queue slot
	int queueOpenCount;	// this is used for tracking the state of the queue

	UndoBuffer(int size) {
		queues = new UndoQueue[size];
		for (int i=0; i<size; i++) {
			queues[i] = new UndoQueue();
		}
		firstQueueIndex = 0;
		lastQueue = -1;
		currentQueue = -1;
		queueArraySize = size;
		queueOpenCount = 0;
	}

	// Opens the queue at currentQueue+1. Makes lastQueue to be the currentQueue,
	// therefore disables redo()
	public void openQueue() {
		if (queueOpenCount > 0) {
			queueOpenCount++;
			return;
		}

		currentQueue++;
		if (currentQueue >= queueArraySize) {
			currentQueue = queueArraySize - 1;
			firstQueueIndex++;
			firstQueueIndex %= queueArraySize;
		}
		lastQueue = currentQueue;
		currentQueueIndex = (firstQueueIndex + currentQueue) % queueArraySize;
		queues[currentQueueIndex].clear();
		queueOpenCount = 1;
	}

	// Closes the current queue
	public void closeQueue() {
		if (queueOpenCount > 0)
			queueOpenCount--;		
	}

	public void storeToQueue(UndoBufferCommand cmd) {
		queues[currentQueueIndex].add(cmd);
	}

	public void undo() {
		if (queueOpenCount > 0 || currentQueue < 0)
			return;

/*		System.out.println("UNDO");
		System.out.println("currentQueue: "+currentQueue);
		System.out.println("currentQueueIndex: "+currentQueueIndex);
		System.out.println("firstQueueIndex: "+firstQueueIndex);
		System.out.println("lastQueue: "+lastQueue);
*/
		queues[currentQueueIndex].undo();
		currentQueue--;
		currentQueueIndex = (firstQueueIndex + currentQueue) % queueArraySize;
	}

	public void redo() {
		if (queueOpenCount > 0 || currentQueue >= lastQueue)
			return;

		currentQueue++;
		currentQueueIndex = (firstQueueIndex + currentQueue) % queueArraySize;

/*		System.out.println("REDO");
		System.out.println("currentQueue: "+currentQueue);
		System.out.println("currentQueueIndex: "+currentQueueIndex);
		System.out.println("firstQueueIndex: "+firstQueueIndex);
		System.out.println("lastQueue: "+lastQueue);
*/
		queues[currentQueueIndex].redo();
	}

	public void clear() {
		firstQueueIndex = 0;
		lastQueue = -1;
		currentQueue = -1;
		queueOpenCount = 0;
	}
}

class UndoQueue {
	Vector queue;

	UndoQueue() {
		queue = new Vector();
	}

	public void add(UndoBufferCommand cmd) {
		queue.addElement(cmd);
	}

	public void undo() {
		int count = queue.size();
//		System.out.println("elements: "+count);
		for (int n=count-1; n>=0; n--) {
			UndoBufferCommand cmd = (UndoBufferCommand)queue.elementAt(n);
			cmd.undo();
		}
	}

	public void redo() {
		int count = queue.size();
//		System.out.println("elements: "+count);
		for (int n=0; n<count; n++) {
			UndoBufferCommand cmd = (UndoBufferCommand)queue.elementAt(n);
			cmd.redo();
		}
	}

	public void clear() {
		queue.removeAllElements();
	}
}

class UndoBufferCommand {
	UndoBufferCommand() {}
	public void undo() {}
	public void redo() {}
}

class VertexBufferState {
	int index;
	Vertex vertex;
	int firstEmptyIdx;
	int lastEmptyIdx;
	Vertex lastEmptyVertex;
	int vertexCount;
	int bufferSize;

	VertexBufferState() {
		vertex = new Vertex();
		lastEmptyVertex = new Vertex();
	}
}

class VertexBufferUndoCmd extends UndoBufferCommand {
	VertexBufferState state;
	VertexBuffer buffer;

	VertexBufferUndoCmd(VertexBuffer vb) {
		state = new VertexBufferState();
		buffer = vb;
	}

	public void undo() {
		buffer.restoreState(state);
	}
}

class VertexBufferRedoCmd extends UndoBufferCommand {
	VertexBufferState state;
	VertexBuffer buffer;

	VertexBufferRedoCmd(VertexBuffer vb) {
		state = new VertexBufferState();
		buffer = vb;
	}

	public void redo() {
		buffer.restoreState(state);
	}
}

class LineBufferState {
	int index;
	Line line;
	int firstEmptyIdx;
	int lastEmptyIdx;
	Line lastEmptyLine;
	int lineCount;
	int bufferSize;
	int previousLine;

	LineBufferState () {
		line = new Line();
		lastEmptyLine = new Line();
	}
}

class LineBufferUndoCmd extends UndoBufferCommand {
	LineBufferState state;
	LineBuffer buffer;

	LineBufferUndoCmd(LineBuffer lb) {
		state = new LineBufferState();
		buffer = lb;
	}

	public void undo() {
		buffer.restoreState(state);
	}
}

class LineBufferRedoCmd extends UndoBufferCommand {
	LineBufferState state;
	LineBuffer buffer;

	LineBufferRedoCmd(LineBuffer lb) {
		state = new LineBufferState();
		buffer = lb;
	}

	public void redo() {
		buffer.restoreState(state);
	}
}

class GridState {
	Matrix4x4 transform;
	float xBend, zBend;
	float cellSize;

	GridState () {
		transform = new Matrix4x4();
	}
}

class GridUndoCmd extends UndoBufferCommand {
	GridState state;
	Grid grid;

	GridUndoCmd(Grid gr) {
		state = new GridState();
		grid = gr;
	}

	public void undo() {
		grid.restoreState(state);
	}
}

class GridRedoCmd extends UndoBufferCommand {
	GridState state;
	Grid grid;

	GridRedoCmd(Grid gr) {
		state = new GridState();
		grid = gr;
	}

	public void redo() {
		grid.restoreState(state);
	}
}
