package v3d.editor;

import java.awt.*;
import java.io.*;

// Vertex class
// - contains Vector3D for position
// - contains reference count & next empty id
class Vertex
{
	public Vector3D vec;
	public int nextEmptyIdx;
	public int refCount;

	Vertex() {
		this.vec = new Vector3D();
	}

	Vertex(Vertex source) {
		copy(source);
	}

	public void copy(Vertex source) {
		this.vec = new Vector3D(source.vec);
		this.nextEmptyIdx = source.nextEmptyIdx;
		this.refCount = source.refCount;
	}

	public void print() {
		vec.print();
		System.out.println("nextEmpty: " + nextEmptyIdx +
					" refCount: " + refCount);
	}
}

// Vertex buffer class
// - add a vertex to the buffer (at the first empty slot)
// - connect a line to a vertex (add a reference)
// - disconnect a line from a vertex (remove a reference)
// - maintain a list of empty vertices in the buffer
// - transform & project the vertex buffer to another buffer
// - resize the vertex buffer automatically
class VertexBuffer
{
	public Vertex vertices[];
	int bufferSize;
	int vertexCount;
	int firstEmptyIdx;
	int lastEmptyIdx;
	UndoBuffer undo = null;

	VertexBuffer(int size) {
		vertices = new Vertex[size];
		for(int i=0; i<size; i++) {
			vertices[i] = new Vertex();
		}
		bufferSize = size;
		clear();
	}

	public synchronized void clear() {
		for(int i=0; i<bufferSize; i++) {
			vertices[i].nextEmptyIdx = i+1;
			vertices[i].refCount = -1;
		}
		vertexCount = 0;
		firstEmptyIdx = 0;
		lastEmptyIdx = bufferSize-1;
	}

	public synchronized int getSize() {
		return bufferSize;
	}

	public synchronized int getVertexCount() {
		return vertexCount;
	}

	// sets the size of the vertex buffer
	void resize(int size) {
		storeUndo(firstEmptyIdx);

		resizeDirect(size);

		//System.out.println("Resize " + bufferSize);
		//System.out.println("firstEmpty " + firstEmptyIdx);
		//System.out.println("firstEmpty.next " + vertices[firstEmptyIdx].nextEmptyIdx);
		//System.out.println("lastEmpty " + vertices[lastEmptyIdx].nextEmptyIdx);

		storeRedo(firstEmptyIdx);
	}

	// Actual implementation for the function above
	void resizeDirect(int size) {
		if (bufferSize < size) {
			VertexBuffer tb = this.duplicate();
			this.vertices = new Vertex[size];
			for(int i=0; i<bufferSize; i++) {
				vertices[i] = new Vertex(tb.vertices[i]);
			}
			for(int i=bufferSize; i<size; i++) {
				vertices[i] = new Vertex();
				vertices[i].nextEmptyIdx = i+1;
				vertices[i].refCount = -1;
			}
			vertices[lastEmptyIdx].nextEmptyIdx = bufferSize;
			lastEmptyIdx = size - 1;
			bufferSize = size;
		}
		else if (bufferSize > size) {
			VertexBuffer tb = this.duplicate();
			this.vertices = new Vertex[size];
			for(int i=0; i<size; i++) {
				vertices[i] = new Vertex(tb.vertices[i]);
			}
			lastEmptyIdx = size - 1;
			bufferSize = size;
		}
	}

	// Duplicates the vertex buffer
	public synchronized VertexBuffer duplicate() {
		VertexBuffer vb = new VertexBuffer(bufferSize);
		for(int i=0; i<bufferSize; i++) {
			vb.vertices[i] = new Vertex(vertices[i]);
		}
		vb.firstEmptyIdx = firstEmptyIdx;
		vb.lastEmptyIdx = lastEmptyIdx;
		vb.vertexCount = vertexCount;
		return vb;
	}

	// Adds a vector to the vertex buffer and adds one reference to the vertex.
	// Returns index of the vertex.
	public synchronized int addVectorWithReference(Vector3D vec) {
		int result = addVector(vec);
		addReference(result);
		return result;
	}

	// Adds a vector to the vertex buffer and returns the index of the vertex.
	// The vector is added to the first empty slot, so a reference should be
	// added right after this if the vertex is used.
	public synchronized int addVector(Vector3D vec) {
		storeUndo(firstEmptyIdx);

		if(firstEmptyIdx == lastEmptyIdx) {
//			System.out.println("add - enlarge");
			this.resize(bufferSize + bufferSize/2 + 1);
		}

		int idx = firstEmptyIdx;
		firstEmptyIdx = vertices[idx].nextEmptyIdx;

		vertices[idx].vec.copy(vec);
		vertices[idx].refCount = 0;

		vertexCount++;

		storeRedo(idx);

		return idx;
	}

	// Adds a reference to a vertex in the buffer. Unreferenced vertices are
	// considered unused.
	public synchronized void addReference(int idx) {
		// it is only possible to reference exisiting vertices
		if(vertices[idx].refCount < 0)
			return;

		storeUndo(idx);

		vertices[idx].refCount += 1;

		storeRedo(idx);
	}

	// Removes a reference from a vertex in the buffer.
	public synchronized void removeReference(int idx) {
		// it is possible to remove reference only from referenced vertices
		if(vertices[idx].refCount < 0)
			return;

		storeUndo(idx);

		vertices[idx].refCount -= 1;

		// It is possible to create vertices that are not in the empty list,
		// but have refCount 0. This removes them as well.
		if(vertices[idx].refCount <= 0)
		{
			vertices[idx].nextEmptyIdx = vertices[lastEmptyIdx].nextEmptyIdx;
			vertices[idx].refCount = -1;
			vertices[lastEmptyIdx].nextEmptyIdx = idx;
			lastEmptyIdx = idx;
			vertexCount--;
		}

		storeRedo(idx);
	}

	public synchronized void setUndoBuffer(UndoBuffer ub) {
		undo = ub;
	}

	void storeUndo(int index) {
		if (undo != null) {
			VertexBufferUndoCmd cmd = new VertexBufferUndoCmd(this);
			storeState(cmd.state,index);
			undo.storeToQueue(cmd);
		}
	}

	void storeRedo(int index) {
		if (undo != null) {
			VertexBufferRedoCmd cmd = new VertexBufferRedoCmd(this);
			storeState(cmd.state,index);
			undo.storeToQueue(cmd);
		}
	}

	// stores the state of the vertex buffer to VertexBufferState
	void storeState(VertexBufferState state, int index) {
		//System.out.println("VertexBuffer.storeState()");
		state.index = index;
		state.vertex.copy(vertices[index]);
		state.firstEmptyIdx = firstEmptyIdx;
		state.lastEmptyIdx = lastEmptyIdx;
		state.lastEmptyVertex.copy(vertices[lastEmptyIdx]);
		state.vertexCount = vertexCount;
		state.bufferSize = bufferSize;
	}

	// restores the state from VertexBufferState
	public synchronized void restoreState(VertexBufferState state) {
		//System.out.println("VertexBuffer.restoreState()");
		if (state.bufferSize != bufferSize) {
			resizeDirect(state.bufferSize);
		}

		vertices[state.index].copy(state.vertex);
		firstEmptyIdx = state.firstEmptyIdx;
		lastEmptyIdx = state.lastEmptyIdx;
		vertices[state.lastEmptyIdx].copy(state.lastEmptyVertex);
		vertexCount = state.vertexCount;
	}

	// Transforms the vertex buffer to another vertex buffer.
	// If the target buffer is smaller than current buffer, it will be
	// resized.
	// This overrides the references and table indexes of the target buffer.
	public synchronized void transform(Matrix4x4 transformMatrix, VertexBuffer target)
	{
		int i;

		if(target.getSize() < bufferSize) {
//			System.out.println("transform");
			target.resize(bufferSize);
		}

		target.firstEmptyIdx = firstEmptyIdx;
		target.lastEmptyIdx = lastEmptyIdx;
		target.vertexCount = vertexCount;

		for(i=0; i<bufferSize; i++) {
			// copy id/reference info
			target.vertices[i].nextEmptyIdx = vertices[i].nextEmptyIdx;
			target.vertices[i].refCount = vertices[i].refCount;

			// transform only if vertex is in use
			if(vertices[i].refCount > 0) {
				transformMatrix.transform(vertices[i].vec,target.vertices[i].vec);
			}
		}

		// The rest of the target buffer is not well-formed after the operation
		// if target buffer is bigger than current buffer, so it has to be cleared
		if (target.bufferSize > bufferSize) {
			for(i=bufferSize; i<target.bufferSize; i++) {
				target.vertices[i].nextEmptyIdx = i+1;
				target.vertices[i].refCount = -1;
			}
			target.lastEmptyIdx = target.bufferSize - 1;
		}
	}

	// Projects the vertex buffer to another vertex buffer with a camera.
	// If the target buffer is smaller than current buffer, it will be
	// resized.
	// This overrides the references and table indexes of the target buffer.
	public synchronized void project(Camera cam, VertexBuffer target)
	{
		int i;

		Vector4D src = new Vector4D();
		Vector4D dst = new Vector4D();

		float halfWidth = cam.viewWidth / 2;
		float halfHeight = cam.viewHeight / 2;

		if(target.getSize() < bufferSize) {
//			System.out.println("project");
			target.resize(bufferSize);
		}

		target.firstEmptyIdx = firstEmptyIdx;
		target.lastEmptyIdx = lastEmptyIdx;
		target.vertexCount = vertexCount;

		for(i=0; i<bufferSize; i++) {
			// copy id/reference info
			target.vertices[i].nextEmptyIdx = vertices[i].nextEmptyIdx;
			target.vertices[i].refCount = vertices[i].refCount;

			// project only if vertex is in use
			if(vertices[i].refCount > 0) {
				src.x = vertices[i].vec.x;
				src.y = vertices[i].vec.y;
				src.z = vertices[i].vec.z;
				src.w = 1;

				cam.projectionMatrix.project(src,dst);

				target.vertices[i].vec.x = halfWidth*(dst.x/dst.w)+cam.viewXOffset+halfWidth;
				target.vertices[i].vec.y = -halfHeight*(dst.y/dst.w)+cam.viewYOffset+halfHeight;
				target.vertices[i].vec.z = dst.z;
			}
		}

		// The rest of the target buffer is not well-formed after the operation
		// if target buffer is bigger than current buffer, so it has to be cleared
		if (target.bufferSize > bufferSize) {
			for(i=bufferSize; i<target.bufferSize; i++) {
				target.vertices[i].nextEmptyIdx = i+1;
				target.vertices[i].refCount = -1;
			}
			target.vertices[lastEmptyIdx].nextEmptyIdx = bufferSize;
			target.lastEmptyIdx = target.bufferSize - 1;
		}
	}

	public synchronized void print() {
		System.out.println("bufferSize: " + bufferSize);
		System.out.println("vertexCount: " + vertexCount);
		System.out.println("firstEmptyIdx: " + firstEmptyIdx);
		System.out.println("lastEmptyIdx: " + lastEmptyIdx);
		for(int i=0; i<bufferSize; i++) {
			if(vertices[i].refCount > 0) {
				vertices[i].print();
			}
		}
	}

	public synchronized int[] compress()
	{
		int[] compressArray = new int[bufferSize];
		int p = 0, n = 0;
		while (p<bufferSize) {
			if (vertices[p].refCount > 0) {
				if (p > n) {
					vertices[n].copy(vertices[p]);
					// tag the vertex to be empty
					vertices[p].refCount = -1;
				}
				compressArray[p] = n;
				n++;
			}
			p++;
		}

		firstEmptyIdx = n;
		lastEmptyIdx = bufferSize - 1;
		for (p = firstEmptyIdx; p <= lastEmptyIdx; p++) {
			vertices[p].nextEmptyIdx = p + 1;
			vertices[p].refCount = -1;
		}
		return compressArray;
	}

	public synchronized void draw(Graphics g, int color) {
		int p;
		g.setColor(new Color(color));
		for (p=0; p<bufferSize; p++) {
			if (vertices[p].refCount > 0) {
				g.drawLine((int)vertices[p].vec.x, (int)vertices[p]. vec.y,
						(int)vertices[p].vec.x, (int)vertices[p].vec.y);
			}
		}
	}

	// Serializes the vertex buffer to a string. Note that the buffer
	// should have been compressed (along with the line buffer) prior to
	// serializing.
	public synchronized String serialize() {
		String s = new String();
		int p;
		for (p=0; p<bufferSize; p++) {
			if (vertices[p].refCount > 0) {
				s += vertices[p].vec.serialize();
			}
		}
		s += "\n";
		return s;
	}

	// Tries to read a vertex from the stream. Returns
	// true if the token was correct and data was read,
	// otherwise false. There is no reference added to the
	// vertex, thus it is assumed that the references will
	// come from loaded lines later.
	public synchronized boolean deserialize(StreamTokenizer st) throws IOException {
		Vector3D vec = new Vector3D();
		if (vec.deserialize(st)) {
			addVector(vec);
			return true;
		}
		else {
			return false;
		}
	}

	public synchronized String getVRML(int indent) {
		String sIndent = new String();
		for (int i=0; i<indent; i++)
			sIndent += "\t";

		String s;
		s = sIndent + "coord Coordinate {\n" +
			sIndent + "\tpoint [\n";

		int p;
		for (p=0; p<bufferSize; p++) {
			if (vertices[p].refCount > 0) {
				s += sIndent + "\t\t" + vertices[p].vec.getVRML() + ",\n";
			}
		}

		s += sIndent + "\t\t]\n" +
			sIndent + "}\n";

		return s;
	}
}
